import os
from .package import PackageSpec
import logging

logger = logging.getLogger(__name__)

PACKAGE_LIST = "packageList.txt"


def find_manifest(package_dir):
    if PACKAGE_LIST in os.listdir(package_dir):
        return os.path.join(package_dir, PACKAGE_LIST)
    return None


def read_manifest(manifest_path):
    """
    Read the list of name requirements and version
    specifications (tags) into a list of lists.
    :param manifest_path: Path to the packageList.txt file
    """
    with open(manifest_path, "r") as pfile:
        return read_manifest_file(pfile)


def read_manifest_file(manifest_file):
    """
    Read the list of name requirements and version
    specifications (tags) into a list of lists.
    :param manifest_file: name file
    :returns List of :py:class:PackageSpec
    """
    package_specs = []
    for line in manifest_file.readlines():
        # Strip comments
        line = line.split("#")[0]
        line = line.rstrip()
        if not len(line):
            continue
        package_specs.append(get_spec(line))
    return package_specs


def update_manifest(manifest_path, package_specs):
    """
    Read the list of name requirements and version
    specifications (tags) into a list of lists.
    :param manifest_path: Path to the packageList.txt file
    :param package_specs: List of package specs to update manifest with
    """
    with open(manifest_path, "r+") as pfile:
        return update_manifest_file(pfile, package_specs)


def update_manifest_file(manifest_file, package_specs):
    """
    Read the list of current requirements and version
    specifications (tags) into a list of lists.
    :param manifest_file: name file
    :param package_specs: List of package specs to update manifest with
    """
    new_specs = {_mangle_spec(spec):spec for spec in package_specs}
    lines = [line for line in manifest_file.readlines()]
    new_lines = []
    for orig in lines:
        # Strip comments
        line_and_comment = orig.split("#")
        line = line_and_comment[0]
        comment = ""
        if len(line_and_comment) > 1:
            comment = "#".join(line_and_comment[1:]).rstrip()
        line = line.rstrip()

        if not len(line):
            # preserve whitespace
            new_lines.append(orig)
            continue

        old_spec = get_spec(line)
        mangled = _mangle_spec(old_spec)
        if mangled in new_specs and old_spec != new_specs[mangled]:
            new_spec = new_specs[mangled]
            line = format_spec(new_spec, comment)
            new_lines.append(line + "\n")
            del new_specs[mangled]
            logger.info("Updating spec for package: {}".format(mangled))
        elif mangled in new_specs:
            new_lines.append(orig)
            del new_specs[mangled]
            logger.info("Package unchanged: {}".format(old_spec.name))

    # Process new specs last
    unused_specs = [i for i in package_specs if i in new_specs.values()]
    for spec in unused_specs:
        logger.info("Writing new spec for package: {}".format(spec.name))
        new_lines.append(format_spec(spec) + "\n")
    manifest_file.seek(0)
    manifest_file.truncate()
    manifest_file.writelines(new_lines)


def _mangle_spec(spec):
    package = spec.name
    if spec.ref_path:
        package = '/'.join([package, spec.ref_path])
    return package


def format_spec(spec, comment=None):
    package = _mangle_spec(spec)
    spec_list = [package, spec.ref]
    if comment:
        spec_list.append(comment)
    return " ".join(spec_list)


def get_spec(line):
    # Rebuild old_spec
    (package, ref) = line.split(" ")
    ref_path = None

    if "/" in package:
        parts = package.split("/")
        package = parts[0]
        ref_path = "/".join(parts[1:])
    return PackageSpec(package, ref, ref_path)



import json
import logging
import os
from .error import RepomanError

logger = logging.getLogger(__name__)

RELEASE_COMMIT_PREFIX = "[repoman-release]"
RELEASE_COMMIT_MESSAGE = "Prepare release"
RELEASE_FILE = "repoman.release.json"
TARGET_DIR = "target"


def resolve_next_version(package, major=None, minor=None, patch=None):
    current_version = package.describe_version()
    split_version = current_version.split("-")
    (next_major, next_minor, next_patch) = [int(i) for i in split_version[:3]]
    pre_release = "-".join(split_version)[3:] if len(split_version) > 3 else ""
    len_args = sum([1 for i in [major, minor, patch] if i])

    if not pre_release and not len_args:
        raise RepomanError("Invalid version specification")

    if major:
        next_major += 1
        next_minor = 0
        next_patch = 0
    elif minor:
        next_minor += 1
        next_patch = 0
    elif patch:
        next_patch += 1

    version = "{:02}-{:02}-{:02}".format(next_major, next_minor, next_patch)
    return version


def prepare(package, release_version, tag_message, commit_message=None,
            tag_dependencies=True, remote="origin"):
    """
    Prepare for a release in git.

    Steps through several phases to ensure the repo is in a sane state
    and the manifest (packageList.txt) is ready to be released,
    resolving files accordingly. After this is done, a release file is
    written and changes are staged for the next step in the release
    process, ``perform``.

    :param package:
    :param release_version: Verion for the package to be released
    :param tag_message: Message for the annotated tag
    :param commit_message: Git Commit message to use after resolving
    the manifest and any other assets. If None, this defaults to
    ``{RELEASE_COMMIT_PREFIX} {RELEASE_COMMIT_MESSAGE} {tag}``, e.g.
    ``[repoman-release] Prepare release astro-01-01-01```
    :param tag_dependencies: If True, tag the dependencies with the
    tag as well.
    :param remote: Remote handle of where to push changes. Defaults
    to ``origin``.
    """
    # Assert everything is committed.
    if package.repo.is_dirty():
        raise RepomanError("Current working copy is dirty. Check git status.")

    current_ref = package.repo.head.commit.hexsha
    new_tag = _get_tag(package, release_version)
    do_resolve_release(package, new_tag, release_version, tag_message)

    full_commit_message = _get_commit_message(commit_message, new_tag)

    release_properties = dict(
        package=package.name,
        tag=new_tag,
        release_version=release_version,
        tag_message=tag_message,
        commit_message=full_commit_message,
        tag_dependencies=tag_dependencies,
        remote=remote,
        current_ref=current_ref
    )

    target_path = os.path.join(package.path, TARGET_DIR)
    release_file_path = os.path.join(target_path, RELEASE_FILE)
    if not os.path.exists(target_path):
        os.mkdir(target_path)
    with open(release_file_path, "w") as release_file:
        output = json.dumps(release_properties)
        release_file.write(output)


def perform(package, push=True):
    """
    Verify state from perform, commit changes, tag package(s),
    and push the tags
    :param package:
    :param push: If True, tagged packages will be pushed.
    """

    target_path = os.path.join(package.path, TARGET_DIR)
    release_file_path = os.path.join(target_path, RELEASE_FILE)

    if not os.path.exists(release_file_path):
        raise RepomanError("No release is currently prepared")

    with open(release_file_path, "r") as release_file:
        release_input = release_file.read()
        release_properties = json.loads(release_input)

    full_commit_message = release_properties["full_commit_message"]
    package.repo.commit(full_commit_message)

    tag = release_properties["tag"]
    tag_message = release_properties["tag_message"]
    remote = release_properties["remote"]
    tag_dependencies = release_properties["tag_dependencies"]

    packages = None
    package.repo.create_tag(tag, ref="HEAD", message=tag_message)

    if package.has_dependencies() and tag_dependencies:
        packages = package.read_manifest()
        # assert_packages_remote(packages)
        seen_dependencies = set()
        for dependency in packages:
            if dependency.name in seen_dependencies:
                logger.warning("Already tagged dependency: {}. "
                               "Skipping...".format(dependency.name))
            ref = dependency.ref
            dependency.repo.create_tag(tag, ref=ref, message=tag_message)
            seen_dependencies.add(dependency.name)

    if push:
        package.repo.remotes[remote].push(tag)
        if packages:
            for dependency in packages:
                dependency.repo.remotes[remote].push(tag)


def do_resolve_release(package, tag, version, tag_message):
    """
    Update the manifest and any other files that might need to be
    updated as part of a release, like release notes,
    documentation, or other.
    :param package:
    :param tag:
    :param version:
    """
    # FIXME: Write this
    # update_manifest(package, tag)
    # release_notes(package, tag, tag_message)
    pass


def _get_tag(package, version):
    return "-".join([package.name, version])


def _get_commit_message(commit_message, tag):
    prefix = RELEASE_COMMIT_PREFIX
    commit_message = commit_message or RELEASE_COMMIT_MESSAGE
    return " ".join([prefix, commit_message, tag])



import os

SCONS_SCRIPTS = ["SConscript", "SConstruct"]


class Tag:

    def __init__(self, package_path):
        self.package_path = package_path

    def tag(self, tag, note, branch=None, custom=None, scons_files=False):
        pass

    def do_git_tag(self, tag, notes, branch=None, release_tag=False, repo_path=None):
        """CvsTag::makeTag"""

        package_file = self._find_package_file()

    def latest_tag(self, filename):
        pass

    def _validate_tag(self, tag, custom):
        pass

    def _find_package_file(self):
        scons_file = None
        for scons in SCONS_SCRIPTS:
            scons_path = os.path.join(self.package_path, scons)
            if os.path.exists(scons_path):
                scons_file = scons_path
        if not scons_file:
            raise FileNotFoundError("No SCons file found for name")
        return scons_file



from git import Repo
from git.exc import GitCommandError
from .error import WorkspaceError
import os
import shutil

DEFAULT_REMOTE_BASE = "git@github.com:fermi-lat"

"""
The Workspace Module is used for staging packages from repositories
to a working directory.
"""


class Workspace:

    def __init__(self, working_path, remote_base=None):
        self.working_path = working_path
        self.remote_base = remote_base or DEFAULT_REMOTE_BASE
        self.repo = None

    def checkout(self, package, ref=None, ref_path=None,
                 force=False, clobber=False):
        """
        Checkout a name repo by name.
        :param package: Name of the name you are checking out
        :param ref: Tag, Branch, or Commit. See `man git-checkout`
        :param ref_path: Specific path to checkout
        :param force: Force git checkout. This throws away local
        changes in the name.
        :param clobber: Remove the directory named `name` first
        """
        repo_path = os.path.join(self.working_path, package)
        # This assumes you aren't using windows

        if clobber:
            if os.path.isdir(repo_path):
                shutil.rmtree(repo_path)

        repo = self.get_or_init_repo(repo_path)
        repo_url = os.path.join(self.remote_base, package) + ".git"
        if not repo.remotes:
            repo.create_remote("origin", repo_url)
        origin = repo.remotes["origin"]

        # Not sure if this needs to be optimized
        try:
            origin.fetch(tags=True)
            origin.fetch()  # This is required for RHEL6/git1.8 support
        except GitCommandError as e:
            raise WorkspaceError("Unable to fetch tags for %s. Please verify "
                                 "name exists and you are accessing it "
                                 "properly. You may also need to wait a few "
                                 "minutes" % package,
                                 "Repo: " + repo_url,
                                 e.stderr)

        checkout_ref = ref or repo.head.ref
        checkout_args = ["-f"] if force else []
        checkout_args.append(checkout_ref)
        if ref_path:
            checkout_args.append(ref_path)
        try:
            repo.git.checkout(*checkout_args)
        except GitCommandError as e:
            raise WorkspaceError("Unable to checkout name: %s, " 
                                 "You may need to force checkout. \n"
                                 "Command Output: " % package,
                                 e.stderr)

    def get_or_init_repo(self, repo_path):
        if os.path.exists(repo_path):
            git_repo = Repo(repo_path)
            if git_repo.working_tree_dir != repo_path:
                raise ValueError("Not in working tree")
        else:
            git_repo = Repo.init(repo_path)
            # Not clear it this does anything, but can't hurt.
            with git_repo.config_writer() as config:
                config.set_value("core", "sparsecheckout", "true")
        return git_repo

    def checkout_packages(self, package_specs, force=False, clobber=False):
        """
        Checkout a bunch of packages
        :param package_specs: list of (name, ref) pairs
        :param force: Force git checkout. This throws away local
        changes in the packages.
        :param clobber: Clobber the name directories
        """
        for spec in package_specs:
            self.checkout(spec.name, spec.ref,
                          spec.ref_path, force=force, clobber=clobber)



from git import Repo
from .error import RepomanError


class Package:
    def __init__(self, name, workspace, path, repo=None):
        self.name = name
        self.workspace = workspace
        self.path = path
        self.repo = repo or Repo(path)
        # FIXME: assert_valid_repo(self.repo)

    def read_manifest(self):
        from .manifest import find_manifest, read_manifest
        manifest_path = find_manifest(self.path)
        if not manifest_path:
            raise RepomanError("Package isn't a product")
        return read_manifest(manifest_path)

    def has_dependencies(self):
        from .manifest import find_manifest
        """
        A product is a special package which is collection of
        packages prepared together into a logical application.

        For Fermi, this includes ScienceTools and GlastRelease.
        """
        return find_manifest(self.path) is not None

    def describe(self):
        """
        Describe this package the most recent package tag.
        """
        # FIXME: Write this
        pass

    def describe_version(self):
        """
        Describe the version string based on the most recent tag.
        """
        # FIXME: Write this
        pass


class PackageSpec:
    def __init__(self, name, ref=None, ref_path=None):
        self.name = name
        self.ref = ref
        self.ref_path = ref_path

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return repr(self.__dict__)



import click
import os
import sys
from .error import RepomanError
from .workspace import Workspace
from .package import Package, PackageSpec
from .manifest import find_manifest, read_manifest, read_manifest_file
from .release import resolve_next_version, prepare, perform


class RepomanCtx(object):

    def __init__(self, workspace_dir, remote_base):
        self.workspace_dir = workspace_dir
        self.remote_base = remote_base
        self.config = {}
        self.verbose = False

    def set_config(self, key, value):
        self.config[key] = value
        if self.verbose:
            click.echo('  config[%s] = %s' % (key, value), file=sys.stderr)

    def __repr__(self):
        return '<RepomanCtx %r>' % self.workspace_dir


pass_ctx = click.make_pass_decorator(RepomanCtx)


@click.group()
@click.option('--workspace', envvar='WORKSPACE_DIR', default=os.getcwd(),
              type=click.Path(),
              metavar='PATH', help='Changes the workspace.')
@click.option('--remote-base', envvar='REMOTE_BASE',
              default="git@github.com:fermi-lat",
              help='Github user/organization for repos')
@click.option('--config', nargs=2, multiple=True,
              metavar='KEY VALUE', help='Overrides a config key/value pair.')
@click.version_option('1.0')
@click.pass_context
def cli(ctx, workspace, remote_base, config):
    """Repoman is a repo and name management tool for
    Fermi's Software configuration.
    """
    # Create a repo object and remember it as as the context object.  From
    # this point onwards other commands can refer to it by using the
    # @pass_repo decorator.
    ctx.obj = RepomanCtx(os.path.abspath(workspace), remote_base)
    for key, value in config:
        ctx.obj.set_config(key, value)


@cli.command()
@click.argument('package')
@click.argument('ref', required=False)
@click.option('--force', is_flag=True,
              help="Force git checkout. This will throw away local changes")
@click.option('--master', is_flag=True,
              help="Ignore versions in name list and check out master")
@pass_ctx
def checkout(ctx, package, ref, force, master):
    """Stage a Fermi package.
    REF may be Tag, Branch, or Commit. For more information,
    see help for git-checkout"""
    workspace = Workspace(ctx.workspace_dir, ctx.remote_base)
    workspace.checkout(package, ref, force=force)
    package_dir = os.path.join(ctx.workspace_dir, package)
    # Check if this is there is a name list
    manifest_path = find_manifest(package_dir)
    if manifest_path is not None:
        package_specs = read_manifest(manifest_path)
        if master:
            package_specs = [PackageSpec(package) for
                             (package, ref, ref_path) in package_specs]
        try:
            workspace.checkout_packages(package_specs)
        except RepomanError as err:
            _print_err(err)
            sys.exit(1)


@cli.command("checkout-list")
@click.argument('package-list', type=click.File("r"))
@click.option('--force', is_flag=True,
              help="Force git checkout. This will throw away local changes")
@click.option('--master', is_flag=True,
              help="Ignore versions in name list and check out master")
@pass_ctx
def checkout_list(ctx, package_list, force, master):
    """Stage packages from a package list."""
    workspace = Workspace(ctx.workspace_dir, ctx.remote_base)
    package_specs = read_manifest_file(package_list)
    if master:
        package_specs = [[package] for (package, ref) in package_specs]
    try:
        workspace.checkout_packages(package_specs, force=force)
    except RepomanError as err:
        _print_err(err)
        sys.exit(1)


@cli.command("release")
@click.argument('package')
@click.argument('release-message')
@click.option('--version',
              help="Custom version for release")
@click.option('--major', is_flag=True,
              help="Bump next major version")
@click.option('--minor', is_flag=True,
              help="Bump next minor version")
@click.option('--patch', is_flag=True,
              help="Bump next patch version")
@click.option('--push/--no-push', default=True,
              help="Push changes")
@pass_ctx
def release(ctx, package, release_message, version, major, minor, patch,
            push):
    """Prepare and perform a release.

    This command executes both the prepare and perform steps
    of a release process.
    """
    package = _get_package(ctx, package)
    if major or minor or patch:
        if version:
            raise RepomanError("Unable to specify version with bumps.")
        version = resolve_next_version(package, major, minor, patch)
    prepare(package, version, release_message)
    perform(package, push=push)


@cli.command("release-prepare")
@click.argument('package')
@click.argument('release-message')
@click.option('--version',
              help="Custom version for release")
@click.option('--major', is_flag=True,
              help="Bump next major version")
@click.option('--minor', is_flag=True,
              help="Bump next minor version")
@click.option('--patch', is_flag=True,
              help="Bump next patch version")
@pass_ctx
def release_prepare(ctx, package, release_message, version,
                    major, minor, patch):
    """Prepare for a release in git.

    Steps through several phases to ensure the repo is in a sane state
    and the manifest (packageList.txt) is ready to be released,
    resolving files accordingly. After this is done, a release file is
    written and changes are staged for the next step in the release
    process, perform.
    """
    package = _get_package(ctx, package)
    if major or minor or patch:
        if version:
            raise RepomanError("Unable to specify version with bumps.")
        version = resolve_next_version(package, major, minor, patch)
    # FIXME: validate_version(package, version)
    prepare(package, version, release_message)


@cli.command("release-perform")
@click.argument('package')
@click.option('--push/--no-push', default=True,
              help="Push changes")
@pass_ctx
def release_perform(ctx, package, push_changes):
    """Perform a release.

    Verify tags and remotes are in order and push them to the
    appropriate remotes."""
    package = _get_package(ctx, package)
    perform(package, push=push_changes)


def _get_package(ctx, name):
    workspace = Workspace(ctx.workspace_dir, ctx.remote_base)
    package_dir = os.path.join(ctx.workspace_dir, name)
    return Package(name, workspace, package_dir)


def _print_err(err):
    for arg in err.args:
        click.echo(arg)


if __name__ == '__main__':
    cli()



"""
Repoman
-------
Repoman is a collection of utilities for managing
individual software packages for the Fermi Gamma-Ray
Space telescope.

Links
`````
* `GitHub <http://github.com/fermi-lat/repoman/>`_
* `development version
  <https://github.com/fermi-lat/repoman/zipball/master#egg=repoman-dev>`_
"""

import re
import ast
from setuptools import setup, find_packages

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('repoman/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


setup(
    name='fermi-repoman',
    version=version,
    packages=find_packages(),
    url='http://github.com/fermi-lat/repoman',
    license='Modified BSD',
    author='Brian Van Klaveren',
    author_email='bvan@slac.stanford.edu',
    description='A collection of utilities for managing '
                'individual software packages for the Fermi Gamma-Ray Space '
                'telescope',
    long_description=__doc__,
    entry_points={
        'console_scripts': ['repoman = repoman.cli:cli'],
    },
    install_requires=[
        'gitpython',
        'click'
    ],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5'
    ],
    zip_safe=True
)
